#!/usr/bin/env python
from test.utils import random_string
from unittest import TestCase
from mock import Mock, MagicMock, patch

from limiter import rate_limit, fungible_limiter, non_fungible_limiter
from limiter.managers import TokenReservation

class BaseLimiterTest(TestCase):
    def setUp(self):
        self.limit = 10
        self.window = 100
        self.token_table_name = random_string()
        self.limit_table_name = random_string()
        self.resource_name = random_string()

class FungibleTokenLimiterDecoratorTest(BaseLimiterTest):
    def setUp(self):
        super(FungibleTokenLimiterDecoratorTest, self).setUp()
        self.mock_manager = Mock()
        self.mock_manager.get_token = Mock()

    def test_call_account_id_pos(self):
        arg_1 = random_string()
        arg_2 = random_string()
        account_id = random_string()
        account_id_pos = 2

        func_to_limit = Mock()
        limiter = rate_limit(
            self.resource_name,
            self.limit,
            self.window,
            self.token_table_name,
            self.limit_table_name,
            account_id_pos)
        limiter._manager = self.mock_manager

        rate_limited_func = limiter.__call__(func_to_limit)
        rate_limited_func(arg_1, arg_2, account_id)

        self.mock_manager.get_token.assert_called_with(account_id)
        func_to_limit.assert_called_with(arg_1, arg_2, account_id)

    def test_call_account_id_key(self):
        arg_1 = random_string()
        arg_2 = random_string()
        account_id = random_string()
        account_id_key = 'account'

        func_to_limit = MagicMock()
        limiter = rate_limit(
            self.resource_name,
            self.limit,
            self.window,
            self.token_table_name,
            self.limit_table_name,
            account_id_key=account_id_key)
        limiter._manager = self.mock_manager

        rate_limited_func = limiter.__call__(func_to_limit)
        rate_limited_func(arg_1, arg_2, account=account_id)

        self.mock_manager.get_token.assert_called_with(account_id)
        func_to_limit.assert_called_with(arg_1, arg_2, account=account_id)

    def test_manager_config_ctor_params(self):
        limiter = rate_limit(
            self.resource_name,
            self.limit,
            self.window,
            self.token_table_name,
            self.limit_table_name)
        manager = limiter.manager

        self.assertEquals(self.token_table_name, manager.token_table_name)
        self.assertEquals(self.limit_table_name, manager.limit_table_name)
        self.assertEquals(self.resource_name, manager.resource_name)

    def test_manager_config_env_params(self):
        env_vars = {
            'FUNGIBLE_TABLE': str(self.token_table_name),
            'LIMIT_TABLE': str(self.limit_table_name)
        }

        with patch.dict('os.environ', env_vars):
            limiter = rate_limit(self.resource_name, self.limit, self.window)
            manager = limiter.manager

            self.assertEquals(self.token_table_name, manager.token_table_name)
            self.assertEquals(self.limit_table_name, manager.limit_table_name)
            self.assertEquals(self.resource_name, manager.resource_name)

    @patch('limiter.limiters.FungibleTokenManager')
    def test_decoratored_account_id_pos(self, mock_manager_delegate):
        arg_1 = random_string()
        arg_2 = random_string()
        account_id = random_string()

        mock_manager = Mock()
        mock_manager.return_value.get_token = Mock()
        mock_manager_delegate.return_value = mock_manager

        self.assertTrue(self._limited_func_account_id_pos(arg_1, arg_2, account_id))
        mock_manager.get_token.assert_called_with(account_id)

    @patch('limiter.limiters.FungibleTokenManager')
    def test_decoratored_account_id_key(self, mock_manager_delegate):
        arg_1 = random_string()
        arg_2 = random_string()
        account_id = random_string()

        mock_manager = Mock()
        mock_manager.return_value.get_token = Mock()
        mock_manager_delegate.return_value = mock_manager

        self.assertTrue(self._limited_func_account_id_key(arg_1, arg_2, account_id=account_id))
        mock_manager.get_token.assert_called_with(account_id)

    @rate_limit('my-resource', 10, 1, 'my-token-table', 'my-limit-table', account_id_pos=3)
    def _limited_func_account_id_pos(self, arg_1, arg_2, account_id):
        self.assertIsNotNone(arg_1)
        self.assertIsNotNone(arg_2)
        self.assertIsNotNone(account_id)
        return True

    @rate_limit('my-resource', 10, 1, 'my-token-table', 'my-limit-table')
    def _limited_func_account_id_key(self, arg_1, arg_2, account_id='my-account'):
        self.assertIsNotNone(arg_1)
        self.assertIsNotNone(arg_2)
        self.assertIsNotNone(account_id)
        return True

class FungibleTokenLimiterContextManagerTest(BaseLimiterTest):
    def setUp(self):
        super(FungibleTokenLimiterContextManagerTest, self).setUp()
        self.account_id = random_string()

    @patch('limiter.limiters.FungibleTokenManager')
    def test_get_token_ctor_params(self, mock_manager_delegate):
        mock_manager = Mock()
        mock_manager.return_value.get_token = Mock()
        mock_manager_delegate.return_value = mock_manager

        with fungible_limiter(self.resource_name, self.account_id, 10, 1, self.token_table_name, self.limit_table_name):
            mock_manager.get_token.assert_called_with(self.account_id)

    @patch('limiter.limiters.FungibleTokenManager')
    def test_get_token_env_params(self, mock_manager_delegate):
        env_vars = {
            'FUNGIBLE_TABLE': str(self.token_table_name),
            'LIMIT_TABLE': str(self.limit_table_name)
        }

        with patch.dict('os.environ', env_vars):
            mock_manager = Mock()
            mock_manager.return_value.get_token = Mock()
            mock_manager_delegate.return_value = mock_manager

            with fungible_limiter(self.resource_name, self.account_id, 10, 1):
                mock_manager.get_token.assert_called_with(self.account_id)

class NonFungibleTokenLimiterContextManagerTest(BaseLimiterTest):
    def setUp(self):
        super(NonFungibleTokenLimiterContextManagerTest, self).setUp()
        self.account_id = random_string()

    @patch('limiter.limiters.NonFungibleTokenManager')
    def test_get_token_ctor_params(self, mock_manager_delegate):
        mock_manager = Mock()
        empty_reservation = TokenReservation(None, None, None, None, None)
        mock_manager.get_reservation = MagicMock(return_value=empty_reservation)
        mock_manager_delegate.return_value = mock_manager

        with non_fungible_limiter(self.resource_name,
                                  self.account_id,
                                  self.limit,
                                  self.token_table_name,
                                  self.limit_table_name) as reservation:
            mock_manager.get_reservation.assert_called_with(self.account_id)
            self.assertIs(empty_reservation, reservation)

    @patch('limiter.limiters.NonFungibleTokenManager')
    def test_get_token_env_params(self, mock_manager_delegate):
        env_vars = {
            'NON_FUNGIBLE_TABLE': self.token_table_name,
            'LIMIT_TABLE': self.limit_table_name
        }

        with patch.dict('os.environ', env_vars):
            mock_manager = Mock()
            empty_reservation = TokenReservation(None, None, None, None, None)
            mock_manager.get_reservation = MagicMock(return_value=empty_reservation)
            mock_manager_delegate.return_value = mock_manager

            with non_fungible_limiter(self.resource_name, self.account_id, self.limit) as reservation:
                mock_manager.get_reservation.assert_called_with(self.account_id)
                self.assertIs(empty_reservation, reservation)

    @patch('limiter.limiters.NonFungibleTokenManager')
    def test_delete_on_exception(self, mock_manager_delegate):
        mock_manager = Mock()
        mock_reservation = Mock()
        mock_reservation.delete = Mock()

        mock_manager.get_reservation = MagicMock(return_value=mock_reservation)
        mock_manager_delegate.return_value = mock_manager

        try:
            with non_fungible_limiter(self.resource_name,
                                      self.account_id,
                                      self.limit,
                                      self.token_table_name,
                                      self.limit_table_name) as reservation:
                self.assertNotNone(reservation)
                raise StandardError('Deliberately thrown from test_delete_on_exception')
        except StandardError:
            pass

        mock_reservation.delete.assert_called()



#!/usr/bin/env python
import random
import string
from datetime import datetime
import boto3

def random_string(length=8):
    return ''.join(random.choice(string.lowercase) for i in range(length))

def now_utc_sec():
    return int(datetime.utcnow().strftime('%s'))

def now_utc_ms():
    return round(float(datetime.utcnow().strftime('%s.%f')) * 1000)

def create_limit_table(table_name, index_name='idx'):
    """ Tests which call this are expected to be in mock_dynamodb2 context """

    mock_client = boto3.client('dynamodb', region_name='us-east-1')
    key_schema = [
        {
            'AttributeName': 'resourceName',
            'KeyType': 'HASH'
        },
        {
            'AttributeName': 'accountId',
            'KeyType': 'RANGE'
        }
    ]

    attribute_definitions = [
        {
            'AttributeName': 'resourceName',
            'AttributeType': 'HASH'
        },
        {
            'AttributeName': 'accountId',
            'AttributeType': 'RANGE'
        },
        {
            'AttributeName': 'limit',
            'AttributeType': 'N'
        },
        {
            'AttributeName': 'windowSec',
            'AttributeType': 'N'
        },
        {
            'AttributeName': 'serviceName',
            'AttributeType': 'S'
        }
    ]

    global_sec_indexes = [
        {
            'IndexName': index_name,
            'KeySchema': [
                {
                    'AttributeName': 'serviceName',
                    'KeyType': 'HASH'
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL'
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 123,
                'WriteCapacityUnits': 123
            }
        }
    ]

    provisioned_throughput = {
        'ReadCapacityUnits': 123,
        'WriteCapacityUnits': 123
    }

    mock_client.create_table(TableName=table_name,
                             KeySchema=key_schema,
                             AttributeDefinitions=attribute_definitions,
                             GlobalSecondaryIndexes=global_sec_indexes,
                             ProvisionedThroughput=provisioned_throughput)

    return boto3.resource('dynamodb', 'us-east-1').Table(table_name)

def create_non_fung_table(table_name, index_name='idx'):
    """ Tests which call this are expected to be in mock_dynamodb2 context """

    mock_client = boto3.client('dynamodb', region_name='us-east-1')
    key_schema = [
        {
            'AttributeName': 'resourceCoordinate',
            'KeyType': 'HASH'
        },
        {
            'AttributeName': 'reservationId',
            'KeyType': 'RANGE'
        }
    ]

    attribute_definitions = [
        {
            'AttributeName': 'resourceCoordinate',
            'AttributeType': 'S'
        },
        {
            'AttributeName': 'reservationId',
            'AttributeType': 'S'
        },
        {
            'AttributeName': 'expirationTime',
            'AttributeType': 'N'
        },
        {
            'AttributeName': 'resourceName',
            'AttributeType': 'S'
        },
        {
            'AttributeName': 'accountId',
            'AttributeType': 'S'
        },
        {
            'AttributeName': 'resourceId',
            'AttributeType': 'S'
        }
    ]

    global_sec_indexes = [
        {
            'IndexName': index_name,
            'KeySchema': [
                {
                    'AttributeName': 'resourceId',
                    'KeyType': 'HASH'
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL'
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 123,
                'WriteCapacityUnits': 123
            }
        }
    ]

    provisioned_throughput = {
        'ReadCapacityUnits': 123,
        'WriteCapacityUnits': 123
    }

    mock_client.create_table(TableName=table_name,
                             KeySchema=key_schema,
                             AttributeDefinitions=attribute_definitions,
                             GlobalSecondaryIndexes=global_sec_indexes,
                             ProvisionedThroughput=provisioned_throughput)

    return boto3.resource('dynamodb', 'us-east-1').Table(table_name)



#!/usr/bin/env python
from limiter.utils import validate_table_env_fallback
from limiter.managers import FungibleTokenManager, NonFungibleTokenManager

class BaseTokenLimiter(object):
    """
    Base class for both fungible and non-fungible token limiters.

    Args:
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
    """
    def __init__(self, resource_name, default_limit):
        self.resource_name = resource_name
        self.default_limit = default_limit
        self._manager = None

class BaseFungibleTokenLimiter(BaseTokenLimiter):
    """
    Extensions of this class provide the most convient usages for rate limiting implementations, e.g. a decorator

    Args:
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
        default_window (str): Sliding window of time, in sec, if no explicit limit is defined in the limits table.
        token_table (str): Name of the DynamoDB table containing tokens.
                           Can be set via environment variable `FUNGIBLE_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        limit_table (str): Name of the DynamoDB table containing limits.
                           Can be set via environment variable `LIMIT_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
    """
    def __init__(self, resource_name, default_limit, default_window, token_table=None, limit_table=None):
        super(BaseFungibleTokenLimiter, self).__init__(resource_name, default_limit)

        self.token_table = validate_table_env_fallback(token_table, 'FUNGIBLE_TABLE', 'fungible-tokens')
        self.limit_table = validate_table_env_fallback(limit_table, 'LIMIT_TABLE', 'limits')
        self.default_window = default_window

    @property
    def manager(self):
        """ Fungible token manager """
        if not self._manager:
            self._manager = FungibleTokenManager(self.token_table,
                                                 self.limit_table,
                                                 self.resource_name,
                                                 self.default_limit,
                                                 self.default_window)
        return self._manager

class FungibleTokenContextManager(BaseFungibleTokenLimiter):
    """
    Fungible token rate-limiter implemented as a context manager.

    Args:
        resource_name (str): Name of the resource being rate-limited.
        account_id (str): The account to retrieve a token on behalf of.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
        default_window (str): Sliding window of time, in sec, if no explicit limit is defined in the limits table.
        token_table (str): Name of the DynamoDB table containing tokens.
                           Can be set via environment variable `FUNGIBLE_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        limit_table (str): Name of the DynamoDB table containing limits.
                           Can be set via environment variable `LIMIT_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.

    Note:
        This class is exported with the more succinct name `fungible_limiter`

    Examples:
        This example assumes `token_table` and `limit_table` have been set via environment variables.

        >>> from limiter import fungible_limiter
        >>> with fungible_limiter('my-resource', 'my-account', 10, 1):
        ...   my_rate_limited_func()
        ...   print 'Done!'
        Done!
    """
    def __init__(self,
                 resource_name,
                 account_id,
                 default_limit,
                 default_window,
                 token_table=None,
                 limit_table=None):
        super(FungibleTokenContextManager, self).__init__(resource_name, default_limit, default_window, token_table,
                                                          limit_table)
        self.account_id = account_id

    def __enter__(self):
        self.get_token()

    def __exit__(self, *args):
        pass

    def get_token(self):
        """ Check the limit and claim a token """
        self.manager.get_token(self.account_id)

class FungibleTokenLimiterDecorator(BaseFungibleTokenLimiter):
    """
    Fungible token rate-limiter implemented as a decorator.

    This decorator requires the account id to be an arguement (positional or keyword) of the function being
    decorated. The location of the account id arguement is set either with `account_id_pos` or `account_id_key`.

    Note:
        This class is exported with the more succinct name `rate_limit`.

    Args:
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
        default_window (str): Sliding window of time, in sec, if no explicit limit is defined in the limits table.
        token_table (str): Name of the DynamoDB table containing tokens.
                           Can be set via environment variable `FUNGIBLE_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        limit_table (str): Name of the DynamoDB table containing limits.
                           Can be set via environment variable `LIMIT_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        account_id_pos (int): Index of the account id in the args of the function being decorated. Defaults to None.
        account_id_key (str): Key of the account id in the kwargs of the function being decorated.
                              Defaults to `account_id`.

    Examples:
        These examples assume `token_table` and `limit_table` have been set via environment variables.

        >>> from limiter import rate_limit
        >>>
        >>> @rate_limit('my-resource', 10, 1, account_id_pos=1)
        ... def first_func(arg_1, account_id)
        ...   print 'In first_func'
        >>>
        >>> @rate_limit('my-resource', 10, 1, account_id_key=my_account_id)
        ... def second_func(arg_1, my_account_id='my-account')
        ...   print 'In second_func'
        >>>
        >>> first_func('foo', 'my-account')
        In first_func
        >>> second_func('bar', 'my-account')
        In second_func
    """
    def __init__(self,
                 resource_name,
                 default_limit,
                 default_window,
                 token_table=None,
                 limit_table=None,
                 account_id_pos=None,
                 account_id_key='account_id'):
        super(FungibleTokenLimiterDecorator, self).__init__(resource_name, default_limit, default_window, token_table,
                                                            limit_table)
        self.is_account_id_kwarg = account_id_pos is None
        self.account_id_index = account_id_key if self.is_account_id_kwarg else account_id_pos

    def __call__(self, func_to_limit):
        def rate_limited_func(*args, **kwargs):
            """
            Extract the account id from the decorated function arguments, fetch a token and call the decorated function.

            Args:
              *args (list): Positional arguments for the decorated function.
              **kwargs (dict): Keyword arguments for the decorated function.

            Returns:
                function: The decorated function wrapped in the account id extraction and token retrieval steps.
            """
            account_id = kwargs[self.account_id_index] if self.is_account_id_kwarg else args[self.account_id_index]
            self.manager.get_token(account_id)
            return func_to_limit(*args, **kwargs)
        return rate_limited_func

class NonFungibleTokenLimiterContextManager(BaseTokenLimiter):
    """
    Non-fungible token rate-limiter implemented as a context manager.

    This class does not create tokens. Rather, it creates instances of TokenReservation which are able to create
    a single token. If the context is exited due to an exception, the token reservation will be deleted and the
    exception propogated.

    Args:
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
        token_table (str): Name of the DynamoDB table containing tokens.
                           Can be set via environment variable `FUNGIBLE_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        limit_table (str): Name of the DynamoDB table containing limits.
                           Can be set via environment variable `LIMIT_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.

    Note:
        This class is exported with the more succinct name `non_fungible_limiter`

    Examples:
        This example assumes `token_table` and `limit_table` have been set via environment variables.

        >>> from limiter import non_fungible_limiter
        >>> with non_fungible_limiter('my-resource', 'my-account', 10) as reservation:
        ...   emr_cluster_id = create_emr_cluster()
        ...   reservation.create_token(emr_cluster_id)
        ...   print 'Done!'
        Done!
    """
    def __init__(self,
                 resource_name,
                 account_id,
                 default_limit,
                 token_table=None,
                 limit_table=None):
        super(NonFungibleTokenLimiterContextManager, self).__init__(resource_name, default_limit)

        self.account_id = account_id
        self.token_table = validate_table_env_fallback(token_table, 'NON_FUNGIBLE_TABLE', 'non-fungible-tokens')
        self.limit_table = validate_table_env_fallback(limit_table, 'LIMIT_TABLE', 'limits')

        self.reservation = None

    @property
    def manager(self):
        """ Non-fungible token manager """
        if not self._manager:
            self._manager = NonFungibleTokenManager(self.token_table, self.limit_table, self.resource_name,
                                                    self.default_limit)
        return self._manager

    def __enter__(self):
        self.reservation = self.get_reservation()
        return self.reservation

    def __exit__(self, *args):
        if any(args):
            print str(args)
            self.reservation.delete()

    def get_reservation(self):
        """ Check the limit and create a reservation """
        return self.manager.get_reservation(self.account_id)

non_fungible_context_manager = NonFungibleTokenLimiterContextManager
fungible_context_manager = FungibleTokenContextManager
decorator = FungibleTokenLimiterDecorator



#!/bin/bash/env python
import logging
import uuid
from datetime import datetime
from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key, Attr
from future.utils import raise_from
from limiter.clients import dynamodb
from limiter.exceptions import CapacityExhaustedException, ReservationNotFoundException, ThrottlingException,\
                               RateLimiterException

logger = logging.getLogger()

# Dynamo table columns
RESOURCE_NAME = 'resourceName'
ACCOUNT_ID = 'accountId'
TOKENS = 'tokens'
LAST_REFILL = 'lastRefill'
LAST_TOKEN = 'lastToken'
RESOURCE_COORDINATE = 'resourceCoordinate'
RESOURCE_ID = 'resourceId'
EXPIRATION_TIME = 'expirationTime'
LIMIT = 'limit'
WINDOW_SEC = 'windowSec'
RESERVATION_ID = 'reservationId'

class BaseTokenManager(object):
    """
    Base class for both fungible and non-fungible token managers.

    Args:
        token_table (str): Name of the DynamoDB table containing tokens.
        limit_table (str): Name of the DynamoDB table containing limits.
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
    """
    def __init__(self, token_table, limit_table, resource_name, default_limit, default_window=0):
        self.token_table_name = token_table
        self.limit_table_name = limit_table
        self.resource_name = resource_name
        self.default_limit = default_limit
        self.default_window = default_window

        self._client = None
        self._token_table = None
        self._limit_table = None

    @property
    def client(self):
        """ DynamoDB client """
        if not self._client:
            self._client = dynamodb()
        return self._client

    @property
    def token_table(self):
        """ DynamoDB Table containing token row """
        if not self._token_table:
            self._token_table = self.client.Table(self.token_table_name)
        return self._token_table

    @property
    def limit_table(self):
        """ DynamoDB Table containing resource limits """
        if not self._limit_table:
            self._limit_table = self.client.Table(self.limit_table_name)
        return self._limit_table

    def _get_account_resource_limit(self, account_id):
        """
        Retrieve the limit and window for an account on this resource.

        If no limits for the given account and this resource is found in DynamoDB a default
        limit and window of 1000 and 1 respectively will be returned.
        An account can be blacklisted by setting its limit to 0. If a limit of 0
        is found a CapacityExhaustedException will be thrown.

        Args:
            account_id (str): The account to get the limits of.

        Returns:
            (dict): Contains the limit and window at the `limit` and `windowSec` keys respectively.

        Throws:
            ThrottlingException: If AWS throttled the query to get resource limit.
            CapacityExhaustedException: If the account has no capacity for the resource, i.e. blacklisted.
            RateLimiterException: On any unrecoverable exception thrown when querying for the resource limit.
        """
        result = {LIMIT: self.default_limit, WINDOW_SEC: self.default_window}
        try:
            response = self.limit_table.query(
                KeyConditionExpression=Key(RESOURCE_NAME).eq(self.resource_name) & Key(ACCOUNT_ID).eq(account_id)
            )
            if response['Count']:
                result = response['Items'][0]
        except Exception as e:
            if isinstance(e, ClientError):
                error_code = e.response['Error']['Code']
                if error_code == 'ProvisionedThroughputExceededException' or error_code == 'TooManyRequestsException':
                    message = 'Throttled getting limit on {} for account {}'.format(self.resource_name, account_id)
                    raise_from(ThrottlingException(message), e)
            message = 'Failed to get limit on {} for account {}'.format(self.resource_name, account_id)
            raise_from(RateLimiterException(message), e)

        if result[LIMIT] <= 0:
            message = 'Account {} has not been allocated any capacity for {}'.format(account_id, self.resource_name)
            raise CapacityExhaustedException(message)
        return result

class FungibleTokenManager(BaseTokenManager):
    """
    Consumes, replenishes and enforces limits on fungible tokens for a single resource stored in a DynamoDB table.

    When a token is requested the manager will conditionally decrement the number of tokens, with a failed
    update signaling hitting the rate limit. If the limit is hit, a `CapacityExhaustedException` will be raised.

    One of three conditions must be true for the token count to be decremented.
    1.  The number of tokens is greater than 0.

    2.  The `lastToken` timestamp is less than the current time - number of milliseconds to accumulate a new token.
        This condition is necessary to ensure clients will still be able to
        access resources when tokens fail to be refilled.
        Note, the minimum number of "accumulation" milliseconds is 1.

    3.  The `tokens` column does not exist. This signifies the row, or bucket, has not been created yet.

    After the manager successfully obtains a token it will add back the number of tokens accumulated
    since the last refill. Updates will be conditionally applied based on last refill time, with stale updates failing.

    Note:
        While constructor's window arguement is expressed in seconds, this class internally represents time in
        milliseconds. The schism enables more granular token rates but makes the API easier to use as
        most limits are expressed in seconds.

    Args:
        token_table (str): Name of the DynamoDB table containing tokens.
        limit_table (str): Name of the DynamoDB table containing limits.
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
        default_window (str): Sliding window of time, in sec, if no explicit limit is defined in the limits table.
    """

    def get_token(self, account_id):
        """
        Retrieve a token on behalf of the specified account.

        If the account has reached its limit a `CapacityExhaustedException` will be raised.
        If a token was successfully retrieved, the number of tokens accumulated since the last
        refill will be added back to the balance.

        Note:
            This method does not actually return a token. If no exception is thrown it means
            the request was not rate limited.

        Args:
            account_id (str): The account to retrieve a token on behalf of.
        """

        exec_time = now_utc_ms()

        limit_response = self._get_account_resource_limit(account_id)
        limit = int(limit_response[LIMIT])
        window_ms = int(limit_response[WINDOW_SEC]) * 1000

        token_ms = float(limit) / window_ms # Number of tokens the bucket will accumulate per millisecond
        ms_token = int(max(1, float(window_ms) / limit)) # Number of milliseconds to accumulate a new token

        bucket = self._get_bucket_token(account_id, exec_time, ms_token)
        current_tokens = bucket[TOKENS]

        last_refill = int(bucket.get('last_refill', 0)) # If the row has not been created yet, use 0
        time_since_refill = exec_time - last_refill
        refill_tokens = _compute_refill_amount(current_tokens, time_since_refill, limit, token_ms)

        self._refill_bucket_tokens(account_id, refill_tokens, exec_time)

    def _get_bucket_token(self, account_id, exec_time, ms_token):
        """
        Conditionally retrieve a token from and return the current state of the bucket (table row).

        Args:
          account_id (str): The account to retrieve a token on behalf of.
          exec_time (int): Time, in milliseconds, when the token retrieval started.
          ms_token (int): The number of milliseconds needed to accumulate a single token.

        Returns:
            dict: State of the bucket after removing a token.

        Raises:
            CapacityExhaustedException: If no more tokens can be taken.
            ThrottlingException: If AWS throttled the query to get resource limit.
            RateLimiterException: On any unrecoverable exception thrown when querying for the resource limit.
        """
        try:
            update_exp = 'add {} :dec set {} = :exec_time'.format(TOKENS, LAST_TOKEN)
            condition_exp = '{0} > :min OR {1} < :failsafe OR attribute_not_exists({0})'.format(TOKENS, LAST_TOKEN)
            return self.token_table.update_item(
                Key={
                    RESOURCE_NAME: self.resource_name,
                    ACCOUNT_ID: account_id
                },
                UpdateExpression=update_exp,
                ConditionExpression=condition_exp,
                ExpressionAttributeValues={
                    ':dec': -1,
                    ':min': 0,
                    ':failsafe': exec_time - ms_token,
                    ':exec_time': exec_time
                },
                ReturnValues='ALL_NEW'
            )['Attributes']
        except Exception as e:
            if isinstance(e, ClientError):
                error_code = e.response['Error']['Code']
                if error_code == 'ConditionalCheckFailedException':
                    message = 'Resource capcity exhausted for {}:{}'.format(self.resource_name, account_id)
                    raise CapacityExhaustedException(message)
                elif error_code == 'ProvisionedThroughputExceededException' or error_code == 'TooManyRequestsException':
                    message = 'Throttled by getting limit on {} for account {}'.format(self.resource_name, account_id)
                    raise_from(ThrottlingException(message), e)
            raise

    def _refill_bucket_tokens(self, account_id, tokens, refill_time):
        """
        Update the token balance to include tokens accumulated since the last refill.

        Args:
          account_id (str): Account which owns the bucket to be refilled.
          tokens (int): The new token balance.
          refill_time (int): Time, in milliseconds, when the token retrieval started.
        """
        try:
            update_exp = 'set {} = :tokens, {} = :refill_time'.format(TOKENS, LAST_REFILL)
            condition_exp = '{0} < :refill_time OR attribute_not_exists({0})'.format(LAST_REFILL)
            self.token_table.update_item(
                Key={
                    RESOURCE_NAME: self.resource_name,
                    ACCOUNT_ID: account_id
                },
                UpdateExpression=update_exp,
                ConditionExpression=condition_exp,
                ExpressionAttributeValues={
                    ':tokens': tokens,
                    ':refill_time': refill_time
                },
                ReturnValues='NONE'
            )
        except Exception as e:
            if isinstance(e, ClientError) and e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                logger.warn('Failed to refill tokens for %s:%s, already refilled with more current state',
                            self.resource_name, account_id)
            else:
                logger.exception('Failed to refill tokens for %s:%s', self.resource_name, account_id)

class NonFungibleTokenManager(BaseTokenManager):
    """
    Creates reservations for and enforces limits on non-fungible tokens for a single resource stored in DynamoDB.

    Unlike FungibleTokenManager, this class does not create tokens. Rather, it creates instances of
    TokenReservation, a placeholder capable of creating a fully constructed token. Both reservations and complete
    tokens count towards the resource limit. When a reservation is created, it is given a TTL of 300 seconds,
    the maximum execution time of a lambda.

    Args:
        token_table (str): Name of the DynamoDB table containing tokens.
        limit_table (str): Name of the DynamoDB table containing limits.
        resource_name (str): Name of the resource being rate-limited.
        default_limit (str): Number of tokens to use if no explicit limit is defined in the limits table.
    """

    def get_reservation(self, account_id):
        """
        Check the token limit and create a TokenReservation.

        Args:
            account_id (str): The account to retrieve a token on behalf of.

        Returns:
            TokenReservation: A reservation to create a token.

        Raises:
            CapacityExhaustedException: If the number of tokens and reservations are at the resource limit.
        """
        exec_time = now_utc_sec()

        limit_response = self._get_account_resource_limit(account_id)
        limit = int(limit_response[LIMIT])

        if self._get_token_count(account_id, exec_time) >= limit:
            message = 'Resource capcity exhausted for {}:{}'.format(self.resource_name, account_id)
            raise CapacityExhaustedException(message)

        return self._build_reservation(account_id, exec_time)

    def _get_token_count(self, account_id, exec_time):
        """
        Get the number of tokens as reservations associated with this resource and the specified account.

        Note:
            Items in DynamoDB which have exceeded their TTL may still appear in query results. This necessitates
            filtering on expirationTime.

        Args:
            account_id (str): The account to get the number of active tokens/reservations for this resource.
            exec_time (int): Timestamp, in seconds, when the calling operation started.

        Returns:
            int: The number of tokens/reservations for this account on this resource.
        """
        coordinate = self._buid_coordinate(account_id)
        return self.token_table.query(
            Select='COUNT',
            ConsistentRead=True,
            KeyConditionExpression=Key(RESOURCE_COORDINATE).eq(coordinate),
            FilterExpression=Attr(EXPIRATION_TIME).gt(exec_time)
        )['Count']

    def _build_reservation(self, account_id, exec_time):
        """
        Insert a reservation into DynamoDB and build and instance of TokenReservation.

        Note:
            The reservation will be inserted with a TTL of 300 seconds.

        Args:
            account_id (str): The account to create the reservation on behalf of.
            exec_time (int): Timestamp, in seconds, when the calling operation started.

        Returns:
            TokenReservation: Reservation to create a token on behalf of the specified account on this resource.
        """
        id = str(uuid.uuid4())
        coordinate = self._buid_coordinate(account_id)
        expiration_time = exec_time + 300

        self.token_table.put_item(
            Item={
                RESOURCE_COORDINATE: coordinate,
                RESERVATION_ID: id,
                RESOURCE_NAME: self.resource_name,
                ACCOUNT_ID: account_id,
                RESOURCE_ID: id,
                EXPIRATION_TIME: expiration_time
            }
        )
        return TokenReservation(id, self.token_table, self.resource_name, account_id, coordinate)

    def _buid_coordinate(self, account_id):
        """
        Build a token resource coordinate for this resource and the specified account.

        Args:
            account_id (str): Id of the account to synthesize the coordinate from.

        Returns:
            str: A coordinate value for this resource and the specified account.
        """
        return '{}:{}'.format(self.resource_name, account_id)

class TokenReservation(object):
    """
    Used to represent a temporary placeholder for, and create a non-fungible token, in DynamoDB.

    Instances of this class represent a single placeholder token for a specific resource and account.
    When a fully formed non-fungible token needs to be created from a reservation it will update the entry
    in DynamoDB with the given resource id and extend its TTL.

    Args:
        id (str): Unique id of the reservation. This is used as the resource id until the full token is created.
        table (boto3.Table): Instance of a boto3 DynamoDB table. This table contains the reservation entry.
        resource_name (str): Name of the resource this reservation is on.
        account_id (str): Account this reservation is for.
        coordinate (str): Resource coordinate this reservation is representing.
    """
    def __init__(self, id, table, resource_name, account_id, coordinate):
        self.id = id
        self.table = table
        self.resource_name = resource_name
        self.account_id = account_id
        self.coordinate = coordinate

        self.is_deleted = False
        self.is_token_created = False

    def create_token(self, resource_id, expiration=28800):
        """
        Create a non-fungible token from this reservation.

        The reservation entry in DynamoDB will be updated to use the given resource id as the token resource id
        and extend the TTL.

        Args:
            resource_id (str): Id of the resource this token will represent, e.g. EMR cluster id.
            expiration (int): The token TTL in seconds. Defaults to 28800 (8 hours).

        Raises:
            ValueError: If a token has already been created from this reservation, or this reservation has been deleted.
            ReservationNotFoundException: If the reservation cannot be found in DynamoDB, likely meaning it expired.
        """
        if self.is_token_created:
            raise ValueError('Token already created for {} from this reservation [{}]'.format(resource_id, self.id))

        if self.is_deleted:
            raise ValueError('This reservation [{}] has been deleted'.format(self.id))

        try:
            expiration_time = now_utc_sec() + expiration
            update_exp = 'set {} = :exp_time, {} = :resource_id'.format(EXPIRATION_TIME, RESOURCE_ID)
            self.table.update_item(
                Key={
                    RESOURCE_COORDINATE: self.coordinate,
                    RESERVATION_ID: self.id
                },
                UpdateExpression=update_exp,
                ExpressionAttributeValues={
                    ':exp_time': expiration_time,
                    ':resource_id': resource_id
                },
                ReturnValues='ALL_NEW'
            )
            self.is_token_created = True
        except ClientError as e:
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                msg_fmt = 'Reservation {} not found for {}:{}. Possibly expired'
                raise ReservationNotFoundException(msg_fmt.format(self.id, self.resource_name, self.account_id))
            raise

    def delete(self):
        """
        Delete the entry in DynamoDB representing this reservation.
        """
        if self.is_token_created:
            logger.warn('Cannot delete, a token has already been created from this reservation [%s]', self.id)
            return

        if self.is_deleted:
            logger.warn('Cannot delete, this reservation [%s], has already been deleted', self.id)
            return

        self.table.delete_item(
            Key={
                RESOURCE_COORDINATE: self.coordinate,
                RESERVATION_ID: self.id
            },
            ReturnValues='NONE'
        )
        self.is_deleted = True

def _compute_refill_amount(current_tokens, time_since_refill, limit, token_ms):
    """
    Compute the number of tokens accumulated since the last refill.

    Args:
      current_tokens (int): The number of tokens currently in the bucket.
      time_since_refill (int): Timestamp, in milliseconds, since the last time the bucket was refilled.
      limit (int): Time, in milliseconds, when the token retrieval started.
      token_ms (int): The rate of token accumulation per millisecond.

    Returns:
        int: The number of tokens accumulated since the last refill.
    """
    tokens = max(0, current_tokens) # Tokens can be negative on bucket creation or a prolonged failure to refill
    return min(limit - 1, tokens + int(token_ms * time_since_refill))

def now_utc_sec():
    """ Get the number of seconds since the epoch """
    return int(datetime.utcnow().strftime('%s'))

def now_utc_ms():
    """ Get the number of milliseconds since the epoch """
    return int(round(float(datetime.utcnow().strftime('%s.%f')) * 1000))



#!/usr/bin/env python
import uuid
import random
from unittest import TestCase
from test.utils import random_string, now_utc_sec, now_utc_ms, create_non_fung_table, create_limit_table
from moto import mock_dynamodb2
from mock import Mock, MagicMock
from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key
from limiter.managers import FungibleTokenManager, NonFungibleTokenManager, TokenReservation, _compute_refill_amount
from limiter.exceptions import CapacityExhaustedException

class FungibleTokenManagerTest(TestCase):
    def setUp(self):
        self.token_table = random_string()
        self.limit_table = random_string()
        self.resource_name = random_string()

        self.limit = 10
        self.window = 100
        self.window_ms = self.window * 1000
        self.token_ms = float(self.limit) / self.window_ms
        self.ms_token = float(self.window_ms) / self.limit

        self.manager = FungibleTokenManager(self.token_table, self.limit_table, self.resource_name, self.limit,
                                            self.window)

    def test_compute_refill_amount(self):
        current_tokens = 5
        time_since_refill = 30000

        expected = 8
        actual = _compute_refill_amount(current_tokens, time_since_refill, self.limit, self.token_ms)

        self.assertEquals(expected, actual)

    def test_compute_refill_amount_negative_balance(self):
        current_tokens = -7
        time_since_refill = 30000

        expected = 3
        actual = _compute_refill_amount(current_tokens, time_since_refill, self.limit, self.token_ms)

        self.assertEquals(expected, actual)

    def test_compute_refill_amount_refill_lag(self):
        current_tokens = 0
        time_since_refill = 11500000

        expected = self.limit - 1
        actual = _compute_refill_amount(current_tokens, time_since_refill, self.limit, self.token_ms)
        self.assertEquals(expected, actual)

    def test_get_bucket_token(self):
        account_id = random_string()
        exec_time = now_utc_ms()

        expected = {'tokens': 5, 'last_refill': now_utc_sec()}
        response = {'Attributes': expected}
        mock_token_table = Mock()
        mock_token_table.update_item = MagicMock(return_value=response)
        self.manager._token_table = mock_token_table

        actual = self.manager._get_bucket_token(account_id, exec_time, self.ms_token)
        self.assertEquals(expected, actual)

        actual_args = mock_token_table.update_item.call_args_list
        self.assertEquals(1, len(actual_args))

        expected_args = [
            ({
                'Key': {
                    'resourceName': self.resource_name,
                    'accountId': account_id
                },
                'UpdateExpression': 'add tokens :dec set lastToken = :exec_time',
                'ConditionExpression': 'tokens > :min OR lastToken < :failsafe OR attribute_not_exists(tokens)',
                'ExpressionAttributeValues': {
                    ':dec': -1,
                    ':min': 0,
                    ':failsafe': exec_time - self.ms_token,
                    ':exec_time': exec_time
                },
                'ReturnValues': 'ALL_NEW'
            })
        ]

        self.assertEquals(expected_args, actual_args[0])

    def test_get_bucket_token_exhausted(self):
        exec_time = now_utc_ms()
        account_id = random_string()
        error_response = {
            'Error': {
                'Code': 'ConditionalCheckFailedException'
            }
        }

        mock_token_table = Mock()
        mock_token_table.update_item = MagicMock(side_effect=ClientError(error_response, None))
        self.manager._token_table = mock_token_table

        self.assertRaises(CapacityExhaustedException,
                          self.manager._get_bucket_token, account_id, exec_time, self.ms_token)

    def test_refill_bucket_tokens(self):
        account_id = random_string()
        tokens = 8
        refill_time = now_utc_ms()

        mock_token_table = Mock()
        mock_token_table.update_item = Mock()
        self.manager._token_table = mock_token_table

        self.manager._refill_bucket_tokens(account_id, tokens, refill_time)

        actual_args = mock_token_table.update_item.call_args_list
        self.assertEquals(1, len(actual_args))

        expected_args = [
            ({
                'Key': {
                    'resourceName': self.resource_name,
                    'accountId': account_id
                },
                'UpdateExpression': 'set tokens = :tokens, lastRefill = :refill_time',
                'ConditionExpression': 'lastRefill < :refill_time OR attribute_not_exists(lastRefill)',
                'ExpressionAttributeValues': {
                    ':tokens': tokens,
                    ':refill_time': refill_time
                },
                'ReturnValues': 'NONE'
            })
        ]

        self.assertEquals(expected_args, actual_args[0])

    @mock_dynamodb2
    def test_account_resource_limit(self):
        limit = random.randint(0, 100)
        window = random.randint(0, 100)
        account_id = random_string()

        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, limit, window)

        self.manager._limit_table = mock_limit_table
        result = self.manager._get_account_resource_limit(account_id)

        self.assertEquals(limit, result['limit'])
        self.assertEquals(window, result['windowSec'])

    @mock_dynamodb2
    def test_account_resource_limit_defaults(self):
        account_id = random_string()

        mock_limit_table = create_limit_table(self.limit_table)
        self.manager._limit_table = mock_limit_table

        result = self.manager._get_account_resource_limit(account_id)
        self.assertEquals(self.limit, result['limit'])
        self.assertEquals(self.window, result['windowSec'])

    @mock_dynamodb2
    def test_account_resource_limit_blacklist(self):
        limit = 0
        window = random.randint(0, 100)
        account_id = random_string()

        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, limit, window)
        self.manager._limit_table = mock_limit_table

        self.assertRaises(CapacityExhaustedException, self.manager._get_account_resource_limit, account_id)

class NonFungibleTokenManagerTest(TestCase):
    def setUp(self):
        self.token_table = random_string()
        self.limit_table = random_string()
        self.resource_name = random_string()

        self.limit = 5
        self.manager = NonFungibleTokenManager(self.token_table, self.limit_table, self.resource_name, self.limit)

    @mock_dynamodb2
    def test_get_reservation(self):
        now = now_utc_sec()
        account_id = random_string()
        coordinate = '{}:{}'.format(self.resource_name, account_id)

        mock_token_table = create_non_fung_table(self.token_table)
        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, self.limit)

        self.manager._token_table = mock_token_table
        self.manager._limit_table = mock_limit_table

        self.manager.get_reservation(account_id)

        response = mock_token_table.query(KeyConditionExpression=Key('resourceCoordinate').eq(coordinate))
        self.assertEquals(1, response['Count'])

        items = dict(pair for item in response['Items'] for pair in item.items())
        self.assertEquals(self.resource_name, items['resourceName'])
        self.assertEquals(account_id, items['accountId'])
        self.assertTrue(items['expirationTime'] > now)
        self.assertIn('resourceId', items)

    @mock_dynamodb2
    def test_get_reservation_exhausted(self):
        now = now_utc_sec()
        account_id = random_string()
        coordinate = '{}:{}'.format(self.resource_name, account_id)

        mock_token_table = create_non_fung_table(self.token_table)
        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, self.limit)

        self.manager._token_table = mock_token_table
        self.manager._limit_table = mock_limit_table

        # Insert enough tokens to reach the limit
        for i in range(0, self.limit):
            token = {
                'resourceCoordinate': coordinate,
                'resourceName': self.resource_name,
                'accountId': account_id,
                'resourceId': 'resource-' + str(i),
                'expirationTime': now + 10000,
                'reservationId': random_string()
            }
            mock_token_table.put_item(Item=token)

        self.assertRaises(CapacityExhaustedException, self.manager.get_reservation, account_id)

    @mock_dynamodb2
    def test_get_token_count(self):
        now = now_utc_sec()
        account_id = random_string()
        coordinate = '{}:{}'.format(self.resource_name, account_id)

        mock_token_table = create_non_fung_table(self.token_table)
        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, self.limit)


        expected_count = 1

        # Insert 2 expired and 1 valid
        expired_token_1 = {
            'resourceCoordinate': coordinate,
            'resourceName': self.resource_name,
            'accountId': account_id,
            'resourceId': random_string(),
            'expirationTime': now,
            'reservationId': random_string()
        }

        expired_token_2 = {
            'resourceCoordinate': coordinate,
            'resourceName': self.resource_name,
            'accountId': account_id,
            'resourceId': random_string(),
            'expirationTime': now - 1000,
            'reservationId': random_string()
        }

        valid_token = {
            'resourceCoordinate': coordinate,
            'resourceName': self.resource_name,
            'accountId': account_id,
            'resourceId': random_string(),
            'expirationTime': now + 300,
            'reservationId': random_string()
        }

        mock_token_table.put_item(Item=expired_token_1)
        mock_token_table.put_item(Item=expired_token_2)
        mock_token_table.put_item(Item=valid_token)

        self.manager._token_table = mock_token_table
        self.manager._limit_table = mock_limit_table
        actual_count = self.manager._get_token_count(account_id, now)

        self.assertEquals(expected_count, actual_count)

    @mock_dynamodb2
    def test_get_token_count_no_tokens(self):
        now = now_utc_sec()
        account_id = random_string()

        mock_token_table = create_non_fung_table(self.token_table)
        mock_limit_table = create_limit_table(self.limit_table)
        _insert_limit(mock_limit_table, self.resource_name, account_id, self.limit)

        self.manager._token_table = mock_token_table
        self.manager._limit_table = mock_limit_table

        expected_count = 0
        actual_count = self.manager._get_token_count(account_id, now)

        self.assertEquals(expected_count, actual_count)

class TokenReservationTest(TestCase):
    def setUp(self):
        self.table_name = random_string()
        self.res_id = str(uuid.uuid4())
        self.resource_name = random_string()
        self.account_id = random_string()
        self.coordinate = '{}:{}'.format(self.resource_name, self.account_id)

    @mock_dynamodb2
    def test_delete_reservation(self):
        mock_token_table = create_non_fung_table(self.table_name)

        reserve = TokenReservation(self.res_id, mock_token_table, self.resource_name, self.account_id, self.coordinate)
        _insert_reservation(mock_token_table, reserve)

        response = mock_token_table.query(KeyConditionExpression=Key('resourceCoordinate').eq(self.coordinate))
        self.assertEquals(1, response['Count'])

        reserve.delete()

        response = mock_token_table.query(KeyConditionExpression=Key('resourceCoordinate').eq(self.coordinate))
        self.assertEquals(0, response['Count'])

    def test_create_after_delete(self):
        mock_token_table = Mock()
        mock_token_table.delete_item = Mock()
        mock_token_table.update_item = Mock()
        reserve = TokenReservation(self.res_id, mock_token_table, self.resource_name, self.account_id, self.coordinate)

        reserve.delete()
        self.assertRaises(ValueError, reserve.create_token, random_string())

    def test_double_create(self):
        mock_token_table = Mock()
        mock_token_table.delete_item = Mock()
        mock_token_table.update_item = Mock()
        reserve = TokenReservation(self.res_id, mock_token_table, self.resource_name, self.account_id, self.coordinate)

        reserve.create_token(random_string())
        self.assertRaises(ValueError, reserve.create_token, random_string())

def _insert_reservation(mock_token_table, reservation):
    reservation_item = {
        'resourceCoordinate': reservation.coordinate,
        'resourceName': reservation.resource_name,
        'accountId': reservation.account_id,
        'resourceId': reservation.id,
        'reservationId': reservation.id,
        'expirationTime': now_utc_sec() + 300
    }
    mock_token_table.put_item(Item=reservation_item)

def _insert_limit(mock_limit_table, resource_name, account_id, limit, window_sec=0):
    limit_item = {
        'resourceName': resource_name,
        'accountId': account_id,
        'limit': limit,
        'windowSec': window_sec,
        'serviceName': random_string()
    }
    mock_limit_table.put_item(Item=limit_item)



class CapacityExhaustedException(Exception):
    """ Raised when a token is requested but none are available. """
    pass

class ReservationNotFoundException(Exception):
    """ Raised when the query result for a non-fungible token reservation is empty. """
    pass

class ThrottlingException(Exception):
    """ Raised when the limiter is throttled by AWS. """
    pass

class RateLimiterException(Exception):
    """ Raised by a limiter on unrecoverable errors when fetching a token or account limits. """



#!/bin/bash/env python
from unittest import TestCase
from test.utils import random_string, now_utc_ms, now_utc_sec, create_non_fung_table
from moto import mock_dynamodb2
from mock import Mock, MagicMock, patch
from boto3.dynamodb.conditions import Key
from limiter.event_processors import ProcessorPredicate, EventProcessor, EventProcessorManager

class ProcessorPredicateTest(TestCase):
    def test_valid_str_value(self):
        target_value = random_string()
        event = {'detail': {'state': target_value}}

        predicate = ProcessorPredicate('detail.state', lambda state: state == target_value)
        self.assertTrue(predicate.test(event))

    def test_valid_numeric_value(self):
        target_value = now_utc_ms()
        event = {'detail': {'time': target_value}}

        predicate = ProcessorPredicate('detail.time', lambda state: state == target_value)
        self.assertTrue(predicate.test(event))

    def test_invalid_value(self):
        target_value = random_string()
        event = {'detail': {'state': random_string()}}

        predicate = ProcessorPredicate('detail.state', lambda state: state == target_value)
        self.assertFalse(predicate.test(event))

    def test_invalid_path(self):
        target_value = random_string()
        event = {'detail': {'state': target_value}}

        predicate = ProcessorPredicate('detail.states', lambda state: state == target_value)
        self.assertFalse(predicate.test(event))

    def test_and_all_true(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == message_value)
        predicate = ProcessorPredicate('detail.zone', lambda state: state == zone_value)

        predicate.with_and(state_pred).with_and(message_pred)
        self.assertTrue(predicate.test(event))

    def test_and_self_false(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == message_value)
        predicate = ProcessorPredicate('detail.zone', lambda state: state == random_string())

        predicate.with_and(state_pred).with_and(message_pred)
        self.assertFalse(predicate.test(event))

    def test_and_sibling_false(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == message_value)
        predicate = ProcessorPredicate('detail.zone', lambda state: state == random_string())

        predicate.with_and(state_pred).with_and(message_pred)
        self.assertFalse(predicate.test(event))

    def test_or_all_true(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == message_value)
        predicate = ProcessorPredicate('detail.zone', lambda state: state == zone_value)

        predicate.with_or(state_pred).with_or(message_pred)
        self.assertTrue(predicate.test(event))

    def test_or_self_false(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == message_value)
        predicate = ProcessorPredicate('detail.zone', lambda state: state == random_string())

        predicate.with_or(state_pred).with_or(message_pred)
        self.assertTrue(predicate.test(event))

    def test_or_sibling_false(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == state_value)
        message_pred = ProcessorPredicate('detail.message', lambda state: state == random_string())
        predicate = ProcessorPredicate('detail.zone', lambda state: state == zone_value)

        predicate.with_or(state_pred).with_or(message_pred)
        self.assertTrue(predicate.test(event))

    def test_or_all_false(self):
        state_value = random_string()
        message_value = random_string()
        zone_value = random_string()
        event = {'detail': {'state': state_value, 'message': message_value, 'zone': zone_value}}

        state_pred = ProcessorPredicate('detail.state', lambda state: state == random_string())
        message_pred = ProcessorPredicate('detail.message', lambda state: state == random_string())
        predicate = ProcessorPredicate('detail.zone', lambda state: state == random_string())

        predicate.with_or(state_pred).with_or(message_pred)
        self.assertFalse(predicate.test(event))

class EventProcessorTest(TestCase):
    def test_with_predicate(self):
        source = random_string()
        expected_id = random_string()
        event = {'detail': {'id': expected_id}}

        mock_predicate = Mock()
        mock_predicate.test = MagicMock(return_value=True)

        processor = EventProcessor(source, 'detail.id', predicate=mock_predicate)
        actual_id = processor.test_and_get_id(event)

        self.assertEquals(expected_id, actual_id)

    def test_without_predicate(self):
        source = random_string()
        expected_id = random_string()
        event = {'detail': {'id': expected_id}}

        processor = EventProcessor(source, 'detail.id')
        actual_id = processor.test_and_get_id(event)

        self.assertEquals(expected_id, actual_id)

    def test_invalid_path(self):
        source = random_string()
        event = {'detail': {'id': random_string()}}

        mock_predicate = Mock()
        mock_predicate.test = MagicMock(return_value=True)

        processor = EventProcessor(source, 'detail.myid', predicate=mock_predicate)
        event_id = processor.test_and_get_id(event)

        self.assertIsNone(event_id)

    def test_failed_predicate(self):
        source = random_string()
        event = {'detail': {'id': random_string()}}

        mock_predicate = Mock()
        mock_predicate.test = MagicMock(return_value=False)

        processor = EventProcessor(source, 'detail.id', predicate=mock_predicate)
        event_id = processor.test_and_get_id(event)

        self.assertIsNone(event_id)

class EventProcessorManagerTest(TestCase):
    def setUp(self):
        self.table_name = random_string()
        self.index_name = random_string()
        self.coordinate = random_string()
        self.resource_name = random_string()
        self.account_id = random_string()
        self.resource_id = random_string()
        self.reservation_id = random_string()

    def test_no_event_source(self):
        event = {'detail': {'state': 'TERMINATED'}}

        mock_processor = Mock()
        mock_processor.source = random_string()
        mock_processor.type = None

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_processor])
        self.assertRaises(ValueError, manager.process_event, event)

    def test_no_processor(self):
        event_source = random_string()
        event = {'source': event_source}

        mock_processor = Mock()
        mock_processor.source = random_string()
        mock_processor.type = None

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_processor])
        self.assertRaises(ValueError, manager.process_event, event)

    def test_env_params(self):
        env_vars = {
            'NON_FUNGIBLE_TABLE': self.table_name,
            'NON_FUNGIBLE_RES_INDEX': self.index_name
        }

        with patch.dict('os.environ', env_vars):
            manager = EventProcessorManager()
            self.assertEquals(self.table_name, manager.table_name)

    @mock_dynamodb2
    def test_delete_token(self):
        event_source = random_string()
        event = {'source': event_source}

        mock_processor = Mock()
        mock_processor.source = event_source
        mock_processor.type = None
        mock_processor.test_and_get_id = MagicMock(return_value=self.resource_id)

        mock_table = create_non_fung_table(self.table_name, self.index_name)
        self._insert_token(mock_table)
        self.assertEquals(1, self._get_resource_id_count(mock_table))

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_processor])
        manager._table = mock_table
        manager.process_event(event)

        self.assertEquals(0, self._get_resource_id_count(mock_table))

    @mock_dynamodb2
    def test_delete_no_token_for_id(self):
        event_source = random_string()
        event = {'source': event_source}

        mock_processor = Mock()
        mock_processor.source = event_source
        mock_processor.type = None
        mock_processor.test_and_get_id = MagicMock(return_value=random_string())

        mock_table = create_non_fung_table(self.table_name, self.index_name)
        self._insert_token(mock_table)
        self.assertEquals(1, self._get_resource_id_count(mock_table))

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_processor])
        manager._table = mock_table
        manager.process_event(event)

        self.assertEquals(1, self._get_resource_id_count(mock_table))

    @mock_dynamodb2
    def test_delete_no_id_from_processor(self):
        event_source = random_string()
        event = {'source': event_source}

        mock_processor = Mock()
        mock_processor.source = event_source
        mock_processor.type = None
        mock_processor.test_and_get_id = MagicMock(return_value=None)

        mock_table = create_non_fung_table(self.table_name, self.index_name)
        self._insert_token(mock_table)
        self.assertEquals(1, self._get_resource_id_count(mock_table))

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_processor])
        manager._table = mock_table
        manager.process_event(event)

        self.assertEquals(1, self._get_resource_id_count(mock_table))

    @mock_dynamodb2
    def test_delete_on_type(self):
        event_source = random_string()
        detail_type = random_string()
        event = {'source': event_source, 'detail-type': detail_type}

        mock_default_processor = Mock()
        mock_default_processor.source = event_source
        mock_default_processor.type = None
        mock_default_processor.test_and_get_id = MagicMock(side_effect=StandardError('Wrong processor invoked'))

        mock_type_processor = Mock()
        mock_type_processor.source = event_source
        mock_type_processor.type = detail_type
        mock_type_processor.test_and_get_id = MagicMock(return_value=self.resource_id)

        mock_table = create_non_fung_table(self.table_name, self.index_name)
        self._insert_token(mock_table)
        self.assertEquals(1, self._get_resource_id_count(mock_table))

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_default_processor, mock_type_processor])
        manager._table = mock_table
        manager.process_event(event)

        self.assertEquals(0, self._get_resource_id_count(mock_table))

    @mock_dynamodb2
    def test_delete_fallback_no_type(self):
        event_source = random_string()
        detail_type = random_string()
        event = {'source': event_source, 'detail-type': detail_type}

        mock_default_processor = Mock()
        mock_default_processor.source = event_source
        mock_default_processor.type = None
        mock_default_processor.test_and_get_id = MagicMock(return_value=self.resource_id)

        mock_type_processor = Mock()
        mock_type_processor.source = event_source
        mock_type_processor.type = detail_type + random_string()
        mock_type_processor.test_and_get_id = MagicMock(side_effect=StandardError('Wrong processor invoked'))

        mock_table = create_non_fung_table(self.table_name, self.index_name)
        self._insert_token(mock_table)
        self.assertEquals(1, self._get_resource_id_count(mock_table))

        manager = EventProcessorManager(table_name=self.table_name,
                                        index_name=self.index_name,
                                        processors=[mock_default_processor, mock_type_processor])
        manager._table = mock_table
        manager.process_event(event)

        self.assertEquals(0, self._get_resource_id_count(mock_table))

    def _get_resource_id_count(self, mock_table):
        response = mock_table.query(IndexName=self.index_name,
                                    KeyConditionExpression=Key('resourceId').eq(self.resource_id))
        return response['Count']

    def _insert_token(self, mock_table):
        token_item = {
            'resourceCoordinate': self.coordinate,
            'resourceName': self.resource_name,
            'accountId': self.account_id,
            'resourceId': self.resource_id,
            'expirationTime': now_utc_sec() + 300,
            'reservationId': self.reservation_id
        }
        mock_table.put_item(Item=token_item)



#!/usr/bin/env python
import os

TABLE_BASE_ENV_VAR = 'LIMITER_TABLES_BASE_NAME'

def validate_table_env_fallback(table_var, env_var, post_fix):
    """
    Check if the given table value is set, falling back to environment variables if not.

    If the given table value is not None, return it. If the value is None, return the specified
    environment variable value, if set. Finally, check if the LIMITER_TABLES_BASE_NAME is set,
    building the table name if set.

    Args:
        table_var (str): Table name value to check.
        env_var (str): Name of the environment variable containing the table name.
        post_fix (str): String to append to the end of LIMITER_TABLES_BASE_NAME.

    Returns:
        str: Name of the table, either the given value, value and the specified env variable or a synthesized one.

    Throws:
        ValueError: If the given table value is None, the specified environment variable is not set
                    and LIMITER_TABLES_BASE_NAME is not set.
    """
    if table_var:
        return table_var
    if env_var in os.environ:
        return os.environ[env_var]
    if TABLE_BASE_ENV_VAR in os.environ:
        base = os.environ[TABLE_BASE_ENV_VAR]
        base = base if base.endswith('-') else base + '-'
        return base + post_fix
    raise ValueError('Failed to retrieve a valid table name')



#!/usr/bin/env python
import json
from boto3.dynamodb.conditions import Key
from limiter.utils import validate_table_env_fallback
from limiter.clients import dynamodb
from limiter.managers import ACCOUNT_ID, RESOURCE_NAME, LIMIT, WINDOW_SEC

# DynamoDB table columns
SERVICE_NAME = 'serviceName'
CONFIG_VERSION = 'configVersion'

class LimitLoader(object):
    """
    Performs initial limit loading and updates for a specified service.

    Instances of this class will only perform a single successful limit load. Subsequent
    calls to load limits will exit without side effects.

    Args:
        service (str): Name of the service limits are being loaded for.
        limit_table (str): Name of the DynamoDB table containing limits.
                           Can be set via environment variable `LIMIT_TABLE`
                           or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        limit_service_index (str): Name of the DynamoDB limit table service index.
                                   Can be set via environment variable `LIMIT_SERVICE_INDEX`
                                   or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
    """
    def __init__(self, service, limit_table=None, limit_service_index=None):
        self.service = service
        self.limit_table_name = validate_table_env_fallback(limit_table, 'LIMIT_TABLE', 'limits')
        self.service_index_name = validate_table_env_fallback(limit_service_index,
                                                              'LIMIT_SERVICE_INDEX', 'limits-service-index')
        self.limit_table = dynamodb().Table(self.limit_table_name)

        self.is_loaded = False

    def load_limits(self, file_path, force=False):
        """
        Update the limits table entries for the given service with those contained in the specified file.

        First, the latest limits are read from the specified JSON file. Next, the limits table is queried for
        limits for the specified service. Any queried limits which are not in the latset limits will be deleted.
        Any queried limits which do not match the latest limits are updated. Any new limits are published.

        Args:
            service (str): Name of the service limits are being loaded for.
            force (bool): Update the limits even if a successful update has already occurred. Defaults to False.
        """
        if self.is_loaded and not force:
            return

        # Get latest limits
        with open(file_path, 'r') as limits_file:
            limits_json = json.load(limits_file)
        latest_limits = limits_json['limits']

        if not latest_limits:
            self.is_loaded = True
            return

        current_limits_response = self._get_current_limits()

        # Index the latest limits by accound and resource to make diffing easier
        indexed_put_requests = {}
        for limit in latest_limits:
            key, item = self._build_put_item(limit)
            indexed_put_requests[key] = item

        # Update limits
        with self.limit_table.batch_writer() as batch:
            for curr_limit in current_limits_response['Items']:
                account_id = curr_limit[ACCOUNT_ID]
                resource_name = curr_limit[RESOURCE_NAME]
                key = account_id + resource_name

                # Current limit is not in the latest set of limits, delete
                if key not in indexed_put_requests:
                    batch.delete_item(Key={
                        ACCOUNT_ID: account_id,
                        RESOURCE_NAME: resource_name
                    })
                    continue

                # If the limit and window are the same remove from from the dict so its not updated
                put_request = indexed_put_requests[key]
                if curr_limit[LIMIT] == put_request[LIMIT] and curr_limit[WINDOW_SEC] == put_request[WINDOW_SEC]:
                    del indexed_put_requests[key]

            # Update out of sync limits and create new ones
            for item in indexed_put_requests.values():
                batch.put_item(item)

        self.is_loaded = True

    def _get_current_limits(self):
        """
        Fetch the current service limits.

        Returns:
            (dict): Query response containing the current service limits.
        """
        return self.limit_table.query(
            IndexName=self.service_index_name,
            KeyConditionExpression=Key(SERVICE_NAME).eq(self.service)
        )

    def _build_put_item(self, limit):
        """
        Build the limit table item and account/resource key.

        Args:
            limit (dict): Limit to create a table item from. Expected to be pulled from the limits JSON file.

        Returns:
            (tuple): Two element tuple. First element is accound_id + resource_name str. Second is the table item.
        """
        account_id = limit[ACCOUNT_ID]
        resource_name = limit[RESOURCE_NAME]

        item = {
            ACCOUNT_ID: account_id,
            RESOURCE_NAME: resource_name,
            LIMIT: limit[LIMIT],
            WINDOW_SEC: limit.get(WINDOW_SEC, 0),
            SERVICE_NAME: self.service
        }
        key = account_id + resource_name
        return key, item



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(name='rate-limiter-py',
    version='0.2.0',
    description='Rate-limiter module which leverages DynamoDB to enforce resource limits.',
    keywords=['lifeomic', 'dynamodb', 'rate', 'limit'],
    author='Matthew Tieman',
    author_email='mjtieman55@gmail.com',
    url='https://github.com/lifeomic/rate-limiter-py',
    download_url='https://github.com/lifeomic/rate-limiter-py/archive/0.2.0.tar.gz',
    packages=find_packages(),
    license='MIT',
    install_requires=[
        'boto3==1.6.4',
        'botocore==1.9.4',
        'future==0.16.0'
    ]
)



#!/bin/bash/env python
import sys
import logging
from boto3.dynamodb.conditions import Key
from limiter.utils import validate_table_env_fallback
from limiter.clients import dynamodb
from limiter.managers import RESOURCE_COORDINATE, RESOURCE_ID, RESERVATION_ID

logger = logging.getLogger()

class ProcessorPredicate(object):
    """
    Determines if an event matches a specific criteria.

    Instances of this class are used to determine if an event value, at the specified
    path, evaluates to True by the given function. Instances may also be composed of additional
    predicates. The additional predicates are either conjunctive (and) or disjunctive (or).

    Args:
        key (str): Location of the value to evalute with the given function.
        pred_func (callable): Function which accepts a single argument, the value, and returns True or False
        and_preds (list:ProcessorPredicate): List of conjunctive ProcessorPredicates. Defaults to an empty list.
        or_preds (list:ProcessorPredicate): List of disjunctive ProcessorPredicates. Defaults to an empty list.

    Note:
        If an instance is composed of both conjunctive and disjunctive predicates, only the conjunctive
        list will be evaluated.

    Examples:
        >>> event = {'source': 'aws.emr', 'detail': {'clusterId': 'j-1YONHTCP3YZKC', 'state': 'COMPLETED'}}
        >>> predicate = ProcessorPredicate('detail.state', lambda state: state == 'COMPLETED')
        >>> predicate.test(event)
        True
    """
    def __init__(self, key, pred_func, and_preds=None, or_preds=None):
        self.key = key
        self.pred_func = pred_func
        self.and_preds = and_preds if and_preds else []
        self.or_preds = or_preds if or_preds else []

    def with_and(self, add_pred):
        """
        Add a predicate to the conjunctive list.

        Args:
            add_pred (ProcessorPredicate): The predicate to add.

        Returns:
            This instance.
        """
        self.and_preds.append(add_pred)
        return self

    def with_or(self, add_pred):
        """
        Add a predicate to the disjunctive list.

        Args:
            add_pred (ProcessorPredicate): The predicate to add.

        Returns:
            This instance.
        """
        self.or_preds.append(add_pred)
        return self

    def test(self, event):
        """
        Evaluate an event.

        Args:
            event (dict): The event to evaluate.

        Returns:
            bool: True if the event satisfies the predicate's conditions, False otherwise.
        """
        test_value = _reduce_to_path(event, self.key)

        result = self.pred_func(test_value) if test_value else False
        if self.and_preds and result:
            for pred in self.and_preds:
                result &= pred.test(event)
                if not result:
                    break
        elif self.or_preds and not result:
            for pred in self.or_preds:
                result |= pred.test(event)
                if result:
                    break
        return result

class EventProcessor(object):
    """
    Validates and extracts the resource id from events.

    Instances of this class will extract the value at the given id path, if the its predicate evaluates to True
    from the event. If an instance does not have a predicate the id will be returned without testing. If either the
    predicate fails or the id path is invalid None will be returned.

    Args:
        source (str): The origination this processor supports events for, e.g. aws.emr
        id_path (str): Dot delineated path to the resource id value.
        predicate (ProcessorPredicate): Predicate to test the event against before extracting the id. Defaults to None.
        type (str): Event detail-type of the event this porcessor supports. Defaults to None.

    Examples:
        >>> event = {'source': 'aws.emr', 'detail': {'clusterId': 'j-1YONHTCP3YZKC', 'state': 'COMPLETED'}}
        >>> predicate = ProcessorPredicate('detail.state', lambda state: state == 'COMPLETED')
        >>> processor = EventProcessor('aws.emr', 'detail.clusterId', predicate=predicate)
        >>> processor.test_and_get_id(event)
        j-1YONHTCP3YZKC
    """
    def __init__(self, source, id_path, predicate=None, type=None):
        self.source = source
        self.id_path = id_path
        self.predicate = predicate
        self.typ = type

    def test_and_get_id(self, event):
        """
        Test the event and extract its resource id.

        If the id path is invalid or the predicate fails None will be returned.

        Args:
            event (dict): Event to test and extract the resource id of.

        Returns:
            str: Id of the resource the event was triggered.
        """
        return None if self.predicate and not self.predicate.test(event) else _reduce_to_path(event, self.id_path)

class EventProcessorManager(object):
    """
    Removes non-fungible tokens from DynamoDB represented by termination events.

    Instances of this class are composed of multiple EventProcessors. Each processor is mapped to
    a single event source, e.g. aws.emr. The processor associated with the event source will verify
    the event is represented by a token and extract the resource id from the event. The resource id
    will then be used to determine the resource coordinate and delete the token.

    Args:
        table_name (str): Name of the DynamoDB non-fungible token table.
                          Can be set via environment variable `NON_FUNGIBLE_TABLE`
                          or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        index_name (str): Name of the index to query for the resource coordinate using the resource id.
                          Can be set via environment variable `NON_FUNGIBLE_RES_INDEX`
                          or synthesized using the `LIMITER_TABLES_BASE_NAME` environment variable.
        processors (list:EventProcessor): List of event processors. Defaults to None.

    Examples:
        >>> event = {'source': 'aws.emr', 'detail': {'clusterId': 'j-1YONHTCP3YZKC', 'state': 'COMPLETED'}}
        >>> predicate = ProcessorPredicate('detail.state', lambda state: state == 'COMPLETED')
        >>> processor = EventProcessor('aws.emr', 'detail.clusterId', predicate=predicate)
        >>> manager = EventProcessorManager(table_name='table', index_name='idx', processors=[processor])
        >>> manager.process_event(event)
    """
    def __init__(self, table_name=None, index_name=None, processors=None):
        self.processors = {_build_processor_key(x.source, x.type): x for x in processors} if processors else {}
        self.table_name = validate_table_env_fallback(table_name, 'NON_FUNGIBLE_TABLE', 'non-fungible-tokens')
        self.index_name = validate_table_env_fallback(index_name, 'NON_FUNGIBLE_RES_INDEX', 'resource-index')

        self.cache = []

        self._client = None
        self._table = None

    @property
    def client(self):
        """ DynamoDB client """
        if not self._client:
            self._client = dynamodb()
        return self._client

    @property
    def table(self):
        """ DynamoDB Table containing token row """
        if not self._table:
            self._table = self.client.Table(self.table_name)
        return self._table

    def add_processor(self, processor):
        """
        Add a processor to this manager.

        If a processor already exists for the event source, it will be replaced.

        Args:
            processor (EventProcessor): Processor to include when handling events.
        """
        self.processors[processor.source] = processor

    def process_event(self, event):
        """
        Remove the token which represents the resource contained in the event.

        First, the EventProcessor will be selected according to the event source.
        If no processor is found a ValueError will be raised.

        Next, the processor will test the event and extract the resource id.
        If the event is malformed or does not satisfy the processor's testing criteria nothing will be deleted.

        Finally, the resource id is used to find the resource coordinate. The resource coordinate and id form
        the key needed to delete the token.

        Args:
            event (dict): Event containing the resource to remove the token of.

        Raises:
            ValueError: If the event does not contain 'source' if this manager has no processor for the source.
        """
        processor = self._get_processor(event)
        resource_id = processor.test_and_get_id(event)
        if resource_id and resource_id not in self.cache:
            token = self._get_resource_token(resource_id)
            if token:
                logger.info('Removing %s token %s from %s', processor.source, resource_id, self.table_name)
                self.table.delete_item(
                    Key={
                        RESOURCE_COORDINATE: token[RESOURCE_COORDINATE],
                        RESERVATION_ID: token[RESERVATION_ID]
                    }
                )
            else:
                logger.warn('Could not find a token for resoure %s', resource_id)
            self.cache.append(resource_id)

    def _get_processor(self, event):
        """
        Get the appropriate processor for this event.

        Args:
            event (dict): Event to processor is being retrieved for.

        Returns:
            EventProcessor: Processor for the event's source.

        Raises:
            ValueError: If the event does not contain 'source' if this manager has no processor for the source.
        """
        if 'source' not in event:
            raise ValueError('Cannot process event, source is a required field. Event: ' + str(event))

        source = event['source']
        type = event.get('detail-type', None)
        source_type_key = _build_processor_key(source, type)

        processor = self.processors.get(source_type_key, None)
        if not processor:
            processor = self.processors.get(source, None)

        if not processor:
            raise ValueError('No processor for event source: ' + source)
        return processor

    def _get_resource_token(self, resource_id):
        """
        Get the token representing the specified resouce.

        Args:
            resource_id (str): Resource id to get the resource token of.

        Returns:
            (dict): The token representing the specpfied resource.
        """
        response = self.table.query(
            IndexName=self.index_name,
            KeyConditionExpression=Key(RESOURCE_ID).eq(resource_id))
        return None if response['Count'] == 0 else response['Items'][0]

def _reduce_to_path(obj, path):
    """
    Traverses the given object down the specified "." delineated, returning the
    value named by the last path segment.

    Args:
        obj (dict): Dictionary containing the value to extract.
        path (str): Location of the value to extract.

    Examples:
        >>> obj = {'foo': {'bar': {'bat': 'SomethingImportant'}}}
        >>> path = 'foo.bar.bat'
        >>> _reduce_to_path(obj, path)
        SomethingImportant

    Returns:
        str: If the path is valid, otherwise None.
    """
    try:
        if isinstance(path, basestring):
            path = path.split('.')
        return reduce(lambda x, y: x[y], path, obj)
    except Exception:
        sys.exc_clear()
    return None

def _build_processor_key(source, type=None):
    return source + ':' + type.replace(' ', '').lower() if type else source



from .limiters import non_fungible_context_manager, fungible_context_manager, decorator

non_fungible_limiter = non_fungible_context_manager
fungible_limiter = fungible_context_manager
rate_limit = decorator
